package at.tuwien.service;

import at.tuwien.api.identifier.BibliographyTypeDto;
import at.tuwien.api.identifier.IdentifierCreateDto;
import at.tuwien.api.identifier.IdentifierUpdateDto;
import at.tuwien.entities.identifier.Identifier;
import at.tuwien.exception.*;
import org.springframework.core.io.InputStreamResource;
import org.springframework.stereotype.Service;

import java.security.Principal;
import java.util.List;

@Service
public interface IdentifierService {

    /**
     * Finds all identifiers in the metadata database which are not deleted. Optionally, the result can be filtered by database id and/or query id.
     *
     * @param databaseId Optional. The database id.
     * @param queryId    Optional. The query id.
     * @return List of identifiers
     * @throws IdentifierNotFoundException The identifier with this filter conditions could not be found in the metadata database.
     */
    List<Identifier> findAll(Long databaseId, Long queryId) throws IdentifierNotFoundException;

    /**
     * Finds all identifiers in the metadata database which are not deleted and filter by query id.
     *
     * @param databaseId The database id.
     * @param queryId    The query id.
     * @return The identifier, if found.
     * @throws IdentifierNotFoundException No identifier with the query id was found.
     */
    Identifier find(Long databaseId, Long queryId) throws IdentifierNotFoundException;

    /**
     * Finds all identifiers
     *
     * @return List of identifiers.
     */
    List<Identifier> findAll();

    /**
     * Creates a new identifier in the metadata database for a query or database.
     *
     * @param data          The identifier.
     * @param principal     The authorization principal.
     * @param authorization The authorization bearer.
     * @return The created identifier from the metadata database if successful.
     * @throws IdentifierPublishingNotAllowedException The identifier with this visibility could not be created.
     * @throws QueryNotFoundException                  The query with this id (in the data) could not be created.
     * @throws RemoteUnavailableException              The connection to the Query Store could not be established by the database connector.
     * @throws IdentifierAlreadyExistsException        The identifier for this query/database already exists.
     * @throws UserNotFoundException                   The user was not found in the metadata database.
     * @throws DatabaseNotFoundException               The database was not found in the metadata database.
     */
    Identifier create(IdentifierCreateDto data, Principal principal, String authorization)
            throws IdentifierPublishingNotAllowedException, QueryNotFoundException,
            RemoteUnavailableException, IdentifierAlreadyExistsException, UserNotFoundException,
            DatabaseNotFoundException, IdentifierRequestException;

    /**
     * Finds an identifier by given id in the metadata database.
     *
     * @param identifierId The identifier id.
     * @return The identifier, if successful.
     * @throws IdentifierNotFoundException The identifier was not found in the metadata database or was deleted.
     */
    Identifier find(Long identifierId) throws IdentifierNotFoundException;

    /**
     * Export metadata for a identifier
     *
     * @param id The identifier id.
     * @return The export, if successful.
     * @throws IdentifierNotFoundException The identifier was not found in the metadata database or was deleted.
     */
    InputStreamResource exportMetadata(Long id) throws IdentifierNotFoundException;

    /**
     * Export metadata for bibliography for a identifier.
     *
     * @param id    The identifier id.
     * @param style The identifier bibliography style. Optional. Default: APA.
     * @return The export, if successful.
     * @throws IdentifierNotFoundException The identifier was not found in the metadata database or was deleted.
     * @throws IdentifierRequestException  The identifier style was not found.
     */
    String exportBibliography(Long id, BibliographyTypeDto style) throws IdentifierNotFoundException, IdentifierRequestException;

    /**
     * Exports an identifier to XML
     *
     * @param identifierId The identifier id.
     * @return The XML resource, if successful.
     * @throws IdentifierNotFoundException The identifier was not found in the metadata database or was deleted.
     * @throws QueryNotFoundException      The query was not found in the metadata database or was deleted.
     * @throws RemoteUnavailableException  The connection to the Query Store could not be established by the database connector.
     * @throws IdentifierRequestException  The identifier does not allow for exporting.
     */
    InputStreamResource exportResource(Long identifierId)
            throws IdentifierNotFoundException, QueryNotFoundException, RemoteUnavailableException, IdentifierRequestException;

    /**
     * Updated the metadata (only) on the identifier for a given id in the metadata database.
     *
     * @param identifierId The identifier id.
     * @param data         The metadata.
     * @return The updated identifier if successful.
     * @throws IdentifierNotFoundException TThe identifier was not found in the metadata database or was deleted.
     */
    Identifier update(Long identifierId, IdentifierUpdateDto data) throws IdentifierNotFoundException, IdentifierRequestException;

    /**
     * Soft-deletes an identifier for a given id in the metadata database. Does not actually remove the entity from the database, but sets it as deleted.
     *
     * @param identifierId The identifier id.
     * @throws IdentifierNotFoundException The identifier was not found in the metadata database or was deleted.
     */
    void delete(Long identifierId) throws IdentifierNotFoundException;
}
