package at.tuwien.service.impl;

import at.tuwien.api.datacite.DataCiteBody;
import at.tuwien.api.datacite.DataCiteData;
import at.tuwien.api.datacite.doi.DataCiteCreateDoi;
import at.tuwien.api.datacite.doi.DataCiteDoi;
import at.tuwien.api.identifier.BibliographyTypeDto;
import at.tuwien.api.identifier.IdentifierCreateDto;
import at.tuwien.api.identifier.IdentifierUpdateDto;
import at.tuwien.config.DataCiteConfig;
import at.tuwien.config.EndpointConfig;
import at.tuwien.entities.identifier.Identifier;
import at.tuwien.exception.*;
import at.tuwien.mapper.DataCiteMapper;
import at.tuwien.repository.jpa.IdentifierRepository;
import at.tuwien.service.IdentifierService;
import lombok.extern.slf4j.Slf4j;
import org.springframework.boot.web.client.RestTemplateBuilder;
import org.springframework.context.annotation.Primary;
import org.springframework.context.annotation.Profile;
import org.springframework.core.ParameterizedTypeReference;
import org.springframework.core.io.InputStreamResource;
import org.springframework.http.*;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.client.HttpClientErrorException;
import org.springframework.web.client.RestClientException;
import org.springframework.web.client.RestTemplate;
import org.springframework.web.util.DefaultUriBuilderFactory;

import java.security.Principal;
import java.util.List;

@Slf4j
@Primary
@Profile("doi")
@Service
public class DataCiteIdentifierServiceImpl implements IdentifierService {

    private final DataCiteConfig dataCiteConfig;
    private final EndpointConfig endpointConfig;
    private final DataCiteMapper dataCiteMapper;
    private final RestTemplateBuilder restTemplateBuilder;
    private final IdentifierRepository identifierRepository;
    private final IdentifierService identifierService;

    public DataCiteIdentifierServiceImpl(DataCiteConfig dataCiteConfig, DataCiteMapper dataCiteMapper,
                                         RestTemplateBuilder restTemplateBuilder, EndpointConfig endpointConfig,
                                         IdentifierRepository identifierRepository, IdentifierServiceImpl identifierService) {
        this.dataCiteConfig = dataCiteConfig;
        this.dataCiteMapper = dataCiteMapper;
        this.restTemplateBuilder =
                restTemplateBuilder.basicAuthentication(dataCiteConfig.getUsername(), dataCiteConfig.getPassword())
                        .uriTemplateHandler(new DefaultUriBuilderFactory(dataCiteConfig.getUrl()));
        this.endpointConfig = endpointConfig;
        this.identifierRepository = identifierRepository;
        this.identifierService = identifierService;
    }

    @Override
    @Transactional(readOnly = true)
    public List<Identifier> findAll(Long databaseId, Long queryId) throws IdentifierNotFoundException {
        return identifierService.findAll(databaseId, queryId);
    }

    @Override
    @Transactional(readOnly = true)
    public Identifier find(Long databaseId, Long queryId) throws IdentifierNotFoundException {
        return identifierService.find(databaseId, queryId);
    }

    @Override
    @Transactional(readOnly = true)
    public List<Identifier> findAll() {
        return identifierService.findAll();
    }

    @Override
    @Transactional(rollbackFor = {Exception.class})
    public Identifier create(IdentifierCreateDto data, Principal principal, String authorization)
            throws IdentifierPublishingNotAllowedException, QueryNotFoundException, RemoteUnavailableException,
            IdentifierAlreadyExistsException, UserNotFoundException, DatabaseNotFoundException,
            IdentifierRequestException {
        Identifier identifier = identifierService.create(data, principal, authorization);
        RestTemplate restTemplate = restTemplateBuilder.build();

        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        headers.setBasicAuth(dataCiteConfig.getUsername(), dataCiteConfig.getPassword());
        HttpEntity<DataCiteBody<DataCiteCreateDoi>> request = new HttpEntity<>(
                DataCiteBody.<DataCiteCreateDoi>builder()
                        .data(DataCiteData.<DataCiteCreateDoi>builder()
                                .type("dois")
                                .attributes(dataCiteMapper.identifierToDataCiteCreateDoi(identifier,
                                        endpointConfig.getWebsiteUrl() + "/pid/" + identifier.getId(),
                                        dataCiteConfig.getPrefix()))
                                .build())
                        .build(),
                headers
        );

        try {
            ResponseEntity<DataCiteBody<DataCiteDoi>> response = restTemplate.exchange("dois", HttpMethod.POST,
                    request,
                    new ParameterizedTypeReference<>() {
                    }
            );

            if (response.getStatusCode() != HttpStatus.CREATED || response.getBody() == null) {
                log.error("Could not successfully create DOI. Response: {}", response);
                throw new IdentifierRequestException("Could not successfully create DOI.");
            }

            identifier.setDoi(response.getBody().getData().getAttributes().getDoi());
            this.identifierRepository.save(identifier);
        } catch (HttpClientErrorException e) {
            log.error("Invalid DOI metadata.", e);
            throw new IdentifierRequestException("Invalid DOI metadata.", e);
        } catch (RestClientException e) {
            log.error("Could not fulfil request to DataCite server.", e);
            throw new InternalError("Could not fulfil request to DataCite server.", e);
        }

        return identifier;
    }

    @Override
    @Transactional(readOnly = true)
    public Identifier find(Long identifierId) throws IdentifierNotFoundException {
        return identifierService.find(identifierId);
    }

    @Override
    @Transactional(readOnly = true)
    public InputStreamResource exportMetadata(Long id) throws IdentifierNotFoundException {
        return identifierService.exportMetadata(id);
    }

    @Override
    @Transactional(readOnly = true)
    public String exportBibliography(Long id, BibliographyTypeDto style)
            throws IdentifierNotFoundException, IdentifierRequestException {
        return identifierService.exportBibliography(id, style);
    }

    @Override
    @Transactional(readOnly = true)
    public InputStreamResource exportResource(Long identifierId)
            throws IdentifierNotFoundException, QueryNotFoundException, RemoteUnavailableException,
            IdentifierRequestException {
        return identifierService.exportResource(identifierId);
    }

    @Override
    @Transactional(rollbackFor = {Exception.class})
    public Identifier update(Long identifierId, IdentifierUpdateDto data)
            throws IdentifierNotFoundException, IdentifierRequestException {
        Identifier identifier = identifierService.update(identifierId, data);
        if (identifier.getDoi() == null) {
            return identifier;
        }

        RestTemplate restTemplate = restTemplateBuilder.build();

        HttpHeaders headers = new HttpHeaders();
        headers.setContentType(MediaType.APPLICATION_JSON);
        headers.setBasicAuth(dataCiteConfig.getUsername(), dataCiteConfig.getPassword());
        HttpEntity<DataCiteBody<DataCiteCreateDoi>> request = new HttpEntity<>(
                DataCiteBody.<DataCiteCreateDoi>builder()
                        .data(DataCiteData.<DataCiteCreateDoi>builder()
                                .type("dois")
                                .attributes(dataCiteMapper.identifierToDataCiteCreateDoi(identifier,
                                        endpointConfig.getWebsiteUrl() + "/pid/" + identifier.getId(),
                                        dataCiteConfig.getPrefix()))
                                .build())
                        .build(),
                headers
        );

        try {
            ResponseEntity<DataCiteBody<DataCiteDoi>> response = restTemplate.exchange("dois/{doi}", HttpMethod.PUT,
                    request,
                    new ParameterizedTypeReference<>() {
                    },
                    identifier.getDoi()
            );

            if (response.getStatusCode() != HttpStatus.OK || response.getBody() == null) {
                log.error("Could not successfully create DOI. Response: {}", response);
                throw new IdentifierRequestException("Could not successfully create DOI.");
            }

            identifier.setDoi(response.getBody().getData().getAttributes().getDoi());
            this.identifierRepository.save(identifier);
        } catch (HttpClientErrorException e) {
            log.error("Invalid DOI metadata.", e);
            throw new IdentifierRequestException("Invalid DOI metadata.", e);
        } catch (RestClientException e) {
            log.error("Could not fulfil request to DataCite server.", e);
            throw new InternalError("Could not fulfil request to DataCite server.", e);
        }

        return identifier;
    }

    @Override
    @Transactional
    public void delete(Long identifierId) throws IdentifierNotFoundException {
        identifierService.delete(identifierId);
    }

}
