package at.tuwien.entities.database.table.columns;

import at.tuwien.entities.container.image.ContainerImageDate;
import at.tuwien.entities.database.View;
import at.tuwien.entities.database.table.Table;
import at.tuwien.entities.user.User;
import lombok.*;
import org.hibernate.annotations.GenericGenerator;
import org.springframework.data.annotation.CreatedDate;
import org.springframework.data.annotation.LastModifiedDate;
import org.springframework.data.jpa.domain.support.AuditingEntityListener;

import jakarta.persistence.*;;
import java.time.Instant;
import java.util.List;

@Data
@Entity
@Builder(toBuilder=true)
@ToString
@AllArgsConstructor
@NoArgsConstructor
@IdClass(TableColumnKey.class)
@EntityListeners(AuditingEntityListener.class)
@jakarta.persistence.Table(name = "mdb_columns", uniqueConstraints = {
        @UniqueConstraint(columnNames = {"tid", "internalName"})
})
public class TableColumn implements Comparable<TableColumn> {

    @Id
    @EqualsAndHashCode.Include
    @GeneratedValue(generator = "columns-sequence")
    @GenericGenerator(name = "columns-sequence", strategy = "increment")
    @Column(updatable = false, nullable = false)
    private Long id;

    @Id
    @EqualsAndHashCode.Include
    private Long tid;

    @Id
    @EqualsAndHashCode.Include
    private Long cdbid;

    @Column
    private Long dfid;

    @ToString.Exclude
    @OneToOne(fetch = FetchType.LAZY, cascade = CascadeType.MERGE)
    @JoinColumn(name = "dfid", referencedColumnName = "id", insertable = false, updatable = false)
    private ContainerImageDate dateFormat;

    @ToString.Exclude
    @ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.MERGE)
    @JoinColumns({
            @JoinColumn(name = "tid", referencedColumnName = "id", insertable = false, updatable = false),
            @JoinColumn(name = "cdbid", referencedColumnName = "tdbid", insertable = false, updatable = false)
    })
    private Table table;

    @ToString.Exclude
    private transient View view;

    @ToString.Exclude
    @ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.PERSIST)
    @JoinColumns({
            @JoinColumn(name = "createdBy", referencedColumnName = "ID", nullable = false, columnDefinition = "VARCHAR(36)", updatable = false)
    })
    private User creator;

    @Column(name = "cname", nullable = false)
    private String name;

    @Column(name = "auto_generated")
    private Boolean autoGenerated = false;

    @Column(nullable = false)
    private String internalName;

    @Column(nullable = false)
    private Boolean isPrimaryKey = false;

    @Column
    private Integer indexLength;

    private transient String alias;

    @Column(name = "datatype", nullable = false)
    @Enumerated(EnumType.STRING)
    private TableColumnType columnType;

    @Column
    private Integer length;

    @Column(nullable = false)
    private Boolean isNullAllowed = true;

    @ElementCollection
    @CollectionTable(name = "mdb_columns_enums", joinColumns = {
            @JoinColumn(name = "id", insertable = false, updatable = false),
            @JoinColumn(name = "tid", insertable = false, updatable = false),
            @JoinColumn(name = "edbid", insertable = false, updatable = false)
    })
    private List<String> enumValues;

    @Column(nullable = false)
    private Integer ordinalPosition;

    @Column(nullable = false, updatable = false, columnDefinition = "TIMESTAMP")
    @CreatedDate
    private Instant created;

    @ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.PERSIST)
    @JoinTable(name = "mdb_columns_concepts",
            joinColumns = {
                    @JoinColumn(name = "cid", referencedColumnName = "id", insertable = false, updatable = false),
                    @JoinColumn(name = "tid", referencedColumnName = "tid", insertable = false, updatable = false),
                    @JoinColumn(name = "cdbid", referencedColumnName = "cdbid", insertable = false, updatable = false)
            },
            inverseJoinColumns = @JoinColumn(name = "uri", referencedColumnName = "uri"))
    private TableColumnConcept concept;

    @ManyToOne(fetch = FetchType.LAZY, cascade = CascadeType.PERSIST)
    @JoinTable(name = "mdb_columns_units",
            joinColumns = {
                    @JoinColumn(name = "cid", referencedColumnName = "id", insertable = false, updatable = false),
                    @JoinColumn(name = "tid", referencedColumnName = "tid", insertable = false, updatable = false),
                    @JoinColumn(name = "cdbid", referencedColumnName = "cdbid", insertable = false, updatable = false)
            },
            inverseJoinColumns = @JoinColumn(name = "uri", referencedColumnName = "uri"))
    private TableColumnUnit unit;

    @LastModifiedDate
    @Column(columnDefinition = "TIMESTAMP")
    private Instant lastModified;

    @Override
    public int compareTo(TableColumn tableColumn) {
        return Integer.compare(this.ordinalPosition, tableColumn.getOrdinalPosition());
    }

    /**
     * KEEP THIS FUNCTION HERE! IT WILL BREAK CODE!
     * Custom equality function implementation.
     *
     * @param object The other column.
     * @return True if columns are equal, false otherwise
     */
    public boolean equals(Object object) {
        if (object == null) {
            return false;
        }
        if (!(object instanceof final TableColumn other)) {
            return false;
        }
        return this.getId().equals(other.getId()) && this.getTid().equals(other.getTid())
                && this.getCdbid().equals(other.getCdbid());
    }
}
