package at.tuwien.endpoint;

import at.tuwien.api.database.table.TableBriefDto;
import at.tuwien.api.maintenance.BannerMessageBriefDto;
import at.tuwien.api.maintenance.BannerMessageCreateDto;
import at.tuwien.api.maintenance.BannerMessageDto;
import at.tuwien.api.maintenance.BannerMessageUpdateDto;
import at.tuwien.api.user.UserBriefDto;
import at.tuwien.exception.BannerMessageNotFoundException;
import at.tuwien.mapper.BannerMessageMapper;
import at.tuwien.service.BannerMessageService;
import io.swagger.v3.oas.annotations.Operation;
import io.swagger.v3.oas.annotations.media.ArraySchema;
import io.swagger.v3.oas.annotations.media.Content;
import io.swagger.v3.oas.annotations.media.Schema;
import io.swagger.v3.oas.annotations.responses.ApiResponse;
import io.swagger.v3.oas.annotations.responses.ApiResponses;
import jakarta.validation.Valid;
import jakarta.validation.constraints.NotNull;
import lombok.extern.log4j.Log4j2;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.http.HttpStatus;
import org.springframework.http.ResponseEntity;
import org.springframework.security.access.prepost.PreAuthorize;
import org.springframework.web.bind.annotation.*;

import java.util.List;

@Log4j2
@CrossOrigin(origins = "*")
@RestController
@RequestMapping("/api/maintenance")
public class MaintenanceEndpoint {

    private final BannerMessageMapper bannerMessageMapper;
    private final BannerMessageService bannerMessageService;

    @Autowired
    public MaintenanceEndpoint(BannerMessageMapper bannerMessageMapper, BannerMessageService bannerMessageService) {
        this.bannerMessageMapper = bannerMessageMapper;
        this.bannerMessageService = bannerMessageService;
    }

    @GetMapping("/message")
    @Operation(summary = "Find maintenance messages")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "List messages",
                    content = {@Content(
                            mediaType = "application/json",
                            array = @ArraySchema(schema = @Schema(implementation = BannerMessageDto.class)))}),
    })
    public ResponseEntity<List<BannerMessageDto>> list() {
        log.debug("endpoint list active maintenance messages");
        final List<BannerMessageDto> dtos = bannerMessageService.findAll()
                .stream()
                .map(bannerMessageMapper::bannerMessageToBannerMessageDto)
                .toList();
        log.trace("list maintenance messages results in dtos {}", dtos);
        return ResponseEntity.ok(dtos);
    }

    @GetMapping("/message/{id}")
    @Operation(summary = "Find one maintenance message")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "Get messages",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = BannerMessageDto.class))}),
    })
    public ResponseEntity<BannerMessageDto> find(@NotNull @PathVariable("id") Long messageId)
            throws BannerMessageNotFoundException {
        log.debug("endpoint find one maintenance messages");
        final BannerMessageDto dto = bannerMessageMapper.bannerMessageToBannerMessageDto(bannerMessageService.find(messageId));
        log.trace("find one maintenance message results in dto {}", dto);
        return ResponseEntity.ok(dto);
    }

    @GetMapping("/message/active")
    @Operation(summary = "Find active maintenance messages")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "200",
                    description = "List messages",
                    content = {@Content(
                            mediaType = "application/json",
                            array = @ArraySchema(schema = @Schema(implementation = BannerMessageBriefDto.class)))}),
    })
    public ResponseEntity<List<BannerMessageBriefDto>> active() {
        log.debug("endpoint list active maintenance messages");
        final List<BannerMessageBriefDto> dtos = bannerMessageService.getActive()
                .stream()
                .map(bannerMessageMapper::bannerMessageToBannerMessageBriefDto)
                .toList();
        log.trace("list active maintenance messages results in dtos {}", dtos);
        return ResponseEntity.ok(dtos);
    }

    @PostMapping("/message")
    @Operation(summary = "Create maintenance message")
    @PreAuthorize("hasAuthority('create-maintenance-message')")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "201",
                    description = "Created message",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = BannerMessageBriefDto.class))}),
    })
    public ResponseEntity<BannerMessageDto> create(@Valid @RequestBody BannerMessageCreateDto data) {
        log.debug("endpoint create maintenance message, data={}", data);
        final BannerMessageDto dto = bannerMessageMapper.bannerMessageToBannerMessageDto(bannerMessageService.create(data));
        log.trace("create maintenance message results in dto {}", dto);
        return ResponseEntity.status(HttpStatus.CREATED)
                .body(dto);
    }

    @PutMapping("/message/{id}")
    @Operation(summary = "Update maintenance message")
    @PreAuthorize("hasAuthority('update-maintenance-message')")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Updated message",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = BannerMessageBriefDto.class))}),
            @ApiResponse(responseCode = "404",
                    description = "Could not find message",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = BannerMessageNotFoundException.class))}),
    })
    public ResponseEntity<BannerMessageDto> update(@NotNull @PathVariable("id") Long messageId,
                                                        @Valid @RequestBody BannerMessageUpdateDto data)
            throws BannerMessageNotFoundException {
        log.debug("endpoint update maintenance message, messageId={}, data={}", messageId, data);
        final BannerMessageDto dto = bannerMessageMapper.bannerMessageToBannerMessageDto(bannerMessageService.update(messageId, data));
        log.trace("update maintenance message results in dto {}", dto);
        return ResponseEntity.status(HttpStatus.ACCEPTED)
                .body(dto);
    }

    @DeleteMapping("/message/{id}")
    @Operation(summary = "Delete maintenance message")
    @PreAuthorize("hasAuthority('delete-maintenance-message')")
    @ApiResponses(value = {
            @ApiResponse(responseCode = "202",
                    description = "Deleted message",
                    content = {@Content(mediaType = "application/json")}),
            @ApiResponse(responseCode = "404",
                    description = "Could not find message",
                    content = {@Content(
                            mediaType = "application/json",
                            schema = @Schema(implementation = BannerMessageNotFoundException.class))}),
    })
    public ResponseEntity<?> delete(@NotNull @PathVariable("id") Long messageId)
            throws BannerMessageNotFoundException {
        log.debug("endpoint delete maintenance message, messageId={}", messageId);
        bannerMessageService.delete(messageId);
        return ResponseEntity.status(HttpStatus.ACCEPTED)
                .build();
    }

}
